#!/usr/bin/env python3
import sys
import time
import select
import curses
from curses import wrapper

entities = []
grid = []

class Wall:
    def collide(self, ball):
        return False

class Block:
    killed = 0
    total = 0

    def __init__(self, x, y, w, h, c):
        self.x = int(round(x))
        self.y = int(round(y))
        self.w = int(round(w))
        self.h = int(round(h))
        self.fmt = curses.A_BOLD | curses.color_pair(c)
        self.alive = True
        for i in range(self.x, self.x + self.w):
            for j in range(self.y, self.y + self.h):
                grid[j + 1][i + 1] = self
        Block.total += 1

    def collide(self, ball):
        self.alive = False
        for i in range(self.x, self.x + self.w):
            for j in range(self.y, self.y + self.h):
                grid[j + 1][i + 1] = None
        Block.killed += 1
        return False

    def tick(self, win):
        if self.alive:
            for i in range(self.x, self.x + self.w):
                for j in range(self.y, self.y + self.h):
                    win.addch(j, i, curses.ACS_BLOCK, self.fmt)
        return self.alive

class Ball:
    alive = False
    killed = 0

    def __init__(self, x, y, vx, vy):
        self.x = int(round(x))
        self.y = int(round(y))
        self.vx = vx
        self.vy = vy
        Ball.alive = True

    def collide(self, ball):
        return True

    def encounter(self, dx, dy):
        dx = int(round(dx))
        dy = int(round(dy))
        ent = grid[self.y + dy + 1][self.x + dx + 1]
        if ent and not ent.collide(self):
            self.vx -= 2 * dx
            self.vy -= 2 * dy
        return ent

    def tick(self, win):
        while self.y < ship.y:
            if self.encounter((self.vx + self.vy) / 2, (self.vy - self.vx) / 2):
                continue
            if self.encounter((self.vx - self.vy) / 2, (self.vy + self.vx) / 2):
                continue
            if self.encounter(self.vx, self.vy):
                continue
            break
        self.x += self.vx
        self.y += self.vy
        try:
            win.addch(self.y, self.x, 'O')
        except curses.error:
            Ball.alive = False
            Ball.killed += 1
        return Ball.alive

class Ship:
    def __init__(self, x, y):
        self.x = int(round(x))
        self.y = int(round(y))
        self.hw = 10
        self.v = 4
        self.last = 1
        self.update()

    def update(self):
        grid[self.y + 1] = (
            [ None ] * (self.x - self.hw + 1) +
            [ self ] * (self.hw * 2 + 1) +
            [ None ] * (width - self.x - self.hw)
        )

    def collide(self, ball):
        ball.vy = -1
        if ball.x > self.x + self.hw / 2:
            ball.vx = 1
        elif ball.x < self.x - self.hw / 2:
            ball.vx = -1
        return True

    def shift(self, i):
        self.last = i
        self.x += self.v * i
        if self.x - self.hw < 0:
            self.x = self.hw
        elif self.x + self.hw >= width:
            self.x = width - self.hw - 1
        self.update()

    def spawn(self):
        if not Ball.alive:
            entities.append(Ball(self.x, self.y - 1, self.last, -1))

    def tick(self, win):
        if not Ball.alive:
            win.addch(self.y - 1, self.x, 'O')
        win.addch(self.y, self.x - self.hw, curses.ACS_LTEE)
        for i in range(-self.hw + 1, self.hw):
            win.addch(curses.ACS_HLINE)
        win.addch(curses.ACS_RTEE)
        return True

class PowerOverwhelmingException(Exception):
    pass

def main(stdscr):
    global height, width, ship

    for i in range(1, 8):
        curses.init_pair(i, i, 0)
    curses.curs_set(0)
    curses.raw()

    height, width = stdscr.getmaxyx()

    if height < 15 or width < 32:
        raise PowerOverwhelmingException(
            'Your computer is not powerful enough to run "arc anoid". '
            'It must support at least 32 columns and 15 rows of next-gen '
            'full-color 3D graphics.')

    status = curses.newwin(1, width, 0, 0)
    height -= 1
    game = curses.newwin(height, width, 1, 0)
    game.nodelay(1)
    game.keypad(1)

    grid[:] = [ [ None for x in range(width + 2) ] for y in range(height + 2) ]
    wall = Wall()
    for x in range(width + 2):
        grid[0][x] = wall
    for y in range(height + 2):
        grid[y][0] = grid[y][-1] = wall
    ship = Ship(width / 2, height - 5)
    entities.append(ship)

    colors = [ 1, 3, 2, 6, 4, 5 ]
    h = height / 10
    for x in range(1, int(width / 7) - 1):
        for y in range(1, 7):
            entities.append(Block(x * 7,
                                  y * h + x / 2 % 2,
                                  7,
                                  h,
                                  colors[y - 1]))

    while True:
        while select.select([ sys.stdin ], [], [], 0)[0]:
            key = game.getch()
            if key == curses.KEY_LEFT or key == ord('a') or key == ord('A'):
                ship.shift(-1)
            elif key == curses.KEY_RIGHT or key == ord('d') or key == ord('D'):
                ship.shift(1)
            elif key == ord(' '):
                ship.spawn()
            elif key == 0x1b or key == 3 or key == ord('q') or key == ord('Q'):
                return

        game.resize(height, width)
        game.erase()
        entities[:] = [ ent for ent in entities if ent.tick(game) ]

        status.hline(0, 0, curses.ACS_HLINE, width)
        status.addch(0, 2, curses.ACS_RTEE)
        status.addstr(' SCORE: ', curses.A_BOLD | curses.color_pair(4))
        status.addstr('%s/%s ' % (Block.killed, Block.total), curses.A_BOLD)
        status.addch(curses.ACS_VLINE)
        status.addstr(' DEATHS: ', curses.A_BOLD | curses.color_pair(4))

        # See T8693. At the minimum display size, we only have room to render
        # two characters for the death count, so just display "99" if the
        # player has more than 99 deaths.
        display_deaths = Ball.killed
        if (display_deaths > 99):
            display_deaths = 99

        status.addstr('%s ' % display_deaths, curses.A_BOLD)
        status.addch(curses.ACS_LTEE)

        if Block.killed == Block.total:
            message = ' A WINNER IS YOU!! '
            i = int(time.time() / 0.8)
            for x in range(width):
                for y in range(6):
                    game.addch(height / 2 + y - 3 + (x / 8 + i) % 2, x,
                               curses.ACS_BLOCK,
                               curses.A_BOLD | curses.color_pair(colors[y]))
            game.addstr(height / 2, (width - len(message)) / 2, message,
                           curses.A_BOLD | curses.color_pair(7))

        game.refresh()
        status.refresh()
        time.sleep(0.05)

try:
    curses.wrapper(main)
    print ('You destroyed %s blocks out of %s with %s deaths.' %
        (Block.killed, Block.total, Ball.killed))
except PowerOverwhelmingException as e:
    print (e)



from __future__ import absolute_import
import sys

is_python_3 = sys.version_info[0] >= 3

if is_python_3:
  def arc_items(dict):
    return dict.items()
else:
  def arc_items(dict):
    return dict.iteritems()

import os
import json

from mercurial import (
  cmdutil,
  bookmarks,
  bundlerepo,
  error,
  hg,
  i18n,
  node,
  registrar,
)

_ = i18n._
cmdtable = {}
command = registrar.command(cmdtable)

@command(
  b'arc-ls-markers',
  [(b'', b'output', b'',
    _(b'file to output refs to'), _(b'FILE')),
  ] + cmdutil.remoteopts,
  _(b'[--output FILENAME] [SOURCE]'))
def lsmarkers(ui, repo, source=None, **opts):
  """list markers

  Show the current branch heads and bookmarks in the local working copy, or
  a specified path/URL.

  Markers are printed to stdout in JSON.

  (This is an Arcanist extension to Mercurial.)

  Returns 0 if listing the markers succeeds, 1 otherwise.
  """

  if source is None:
    markers = localmarkers(ui, repo)
  else:
    markers = remotemarkers(ui, repo, source, opts)

  for m in markers:
    if m['name'] != None:
      m['name'] = m['name'].decode('utf-8')

    if m['node'] != None:
      m['node'] = m['node'].decode('utf-8')

    if m['description'] != None:
      m['description'] = m['description'].decode('utf-8')

  json_opts = {
    'indent': 2,
    'sort_keys': True,
  }

  output_file = opts.get('output')
  if output_file:
    if os.path.exists(output_file):
      raise error.Abort(_('File "%s" already exists.' % output_file))
    with open(output_file, 'w+') as f:
      json.dump(markers, f, **json_opts)
  else:
    json_data = json.dumps(markers, **json_opts)
    print(json_data)

  return 0

def localmarkers(ui, repo):
  markers = []

  active_node = repo[b'.'].node()
  all_heads = set(repo.heads())
  current_name = repo.dirstate.branch()

  branch_list = repo.branchmap().iterbranches()
  for branch_name, branch_heads, tip_node, is_closed in branch_list:
    for head_node in branch_heads:

      is_active = False
      if branch_name == current_name:
        if head_node == active_node:
          is_active = True

      is_tip = (head_node == tip_node)

      if is_closed:
        head_closed = True
      else:
        head_closed = bool(head_node not in all_heads)

      description = repo[head_node].description()

      markers.append({
        'type': 'branch',
        'name': branch_name,
        'node': node.hex(head_node),
        'isActive': is_active,
        'isClosed': head_closed,
        'isTip': is_tip,
        'description': description,
      })

  bookmarks = repo._bookmarks
  active_bookmark = repo._activebookmark

  for bookmark_name, bookmark_node in arc_items(bookmarks):
    is_active = (active_bookmark == bookmark_name)
    description = repo[bookmark_node].description()

    markers.append({
      'type': 'bookmark',
      'name': bookmark_name,
      'node': node.hex(bookmark_node),
      'isActive': is_active,
      'description': description,
    })

  # Add virtual markers for the current commit state and current branch state
  # so callers can figure out exactly where we are.

  # Common cases where this matters include:

  # You run "hg update 123" to update to an older revision. Your working
  # copy commit will not be a branch head or a bookmark.

  # You run "hg branch X" to create a new branch, but have not made any commits
  # yet. Your working copy branch will not be reflected in any commits.

  markers.append({
    'type': 'branch-state',
    'name': current_name,
    'node': None,
    'isActive': True,
    'isClosed': False,
    'isTip': False,
    'description': None,
  })

  markers.append({
    'type': 'commit-state',
    'name': None,
    'node': node.hex(active_node),
    'isActive': True,
    'isClosed': False,
    'isTip': False,
    'description': repo[b'.'].description(),
  })

  return markers

def remotemarkers(ui, repo, source, opts):
  # Disable status output from fetching a remote.
  ui.quiet = True

  markers = []

  source, branches = hg.parseurl(ui.expandpath(source))
  remote = hg.peer(repo, opts, source)

  with remote.commandexecutor() as e:
    branchmap = e.callcommand(b'branchmap', {}).result()

  for branch_name in branchmap:
    for branch_node in branchmap[branch_name]:
      markers.append({
        'type': 'branch',
        'name': branch_name,
        'node': node.hex(branch_node),
        'description': None,
      })

  with remote.commandexecutor() as e:
    remotemarks = bookmarks.unhexlifybookmarks(e.callcommand(b'listkeys', {
      b'namespace': b'bookmarks',
    }).result())

  for mark in remotemarks:
    markers.append({
      'type': 'bookmark',
      'name': mark,
      'node': node.hex(remotemarks[mark]),
      'description': None,
    })

  return markers
